/* H.323 Call Generator.
 (c) Benny Prijono <seventhson@theseventhson.freeserve.co.uk>
 
 Changelog:

 $Log: main.cxx,v $
 Revision 1.8  2002/05/31 04:32:33  robertj
 Fixed bad conditional, thanks Federico Pinna

 Revision 1.7  2002/05/27 00:30:03  robertj
 Added adjustable time ranges for random calls, thanks Federico Pinna

 Revision 1.6  2002/05/10 00:25:30  robertj
 Changed to trace out data and time as this is more useful for matching up
   with other endpoints or ethereal traces (provided clocks a synchronised).

 Revision 1.5  2002/03/14 05:00:36  robertj
 Added some tracing of call starts.
 Added console output of end of run in each thread.

 Revision 1.4  2002/02/19 06:01:06  rogerh
 Added --listenport option. Submitted by andreas.sikkema@philips.com
 Added history log.


 25 Jan 2002 Substantial improvement [Equivalence Pty. Ltd.]
 25 Jan 2000 Update to incorporate openh323 v.01 alpha2 and fix gatekeeper
             related codes [bennylp]
 */
#include "precompile.h"
#include "main.h"
#include "version.h"


#include <ptclib/pwavfile.h>
#include <ptclib/delaychan.h>
#include <ptclib/random.h>


PCREATE_PROCESS(CallGen323);

PMutex coutMutex;


///////////////////////////////////////////////////////////////////////////////

BOOL COH323EndPoint::Initialise(const PArgList& args)
{
  totalEstablished = 0;

  // set dummy capability
  //SetCapability(0, 0, new COHG711ULaw64kCapability);
  SetCapability(0, 0, new H323_G711Capability(H323_G711Capability::muLaw, H323_G711Capability::At64k));
  H323_UserInputCapability::AddAllCapabilities(capabilities, 0, P_MAX_INDEX);


  WORD listenPort = H323ListenerTCP::DefaultSignalPort;
  H323ListenerTCP * listener;

  if (!args.GetOptionString("listenport").IsEmpty())
  {
    listenPort = (WORD)args.GetOptionString("listenport").AsInteger();
  }

  cout << "ListenPort = " << listenPort << endl;

  if (args.GetOptionString('i').IsEmpty())
  {
    PIPSocket::Address interface_addr(INADDR_ANY);
    listener = new H323ListenerTCP(*this, interface_addr, listenPort, FALSE);
  }
  else
  {
    PIPSocket::Address interface_addr(args.GetOptionString('i'));
    listener = new H323ListenerTCP(*this, interface_addr, listenPort, FALSE);
  }

  // init listener
  if (!StartListener(listener)) 
  {
    cout << "Couldn't start listener on port "
         << listener->GetListenerPort() << endl;
    delete listener;
    return FALSE;
  }
  
  // set local username, is necessary
  if (args.HasOption('u')) {
    PStringArray aliases = args.GetOptionString('u').Lines();
    SetLocalUserName(aliases[0]);
    for (PINDEX i = 1; i < aliases.GetSize(); ++i)
      AddAliasName(aliases[i]);
  }
  cout << "Local username: " << GetLocalUserName() << endl;
  
  // process gatekeeper registration options
  if (args.HasOption('g')) {
    PString gkName = args.GetOptionString('g');
    H323TransportUDP * rasChannel;
    if (args.GetOptionString('i').IsEmpty())
      rasChannel  = new H323TransportUDP(*this);
    else {
      PIPSocket::Address interfaceAddress(args.GetOptionString('i'));
      rasChannel  = new H323TransportUDP(*this, interfaceAddress);
    }
    if (SetGatekeeper("ip$"+gkName, rasChannel))
      cout << "Gatekeeper set: " << *gatekeeper << endl;
    else {
      cout << "Error registering with gatekeeper at \"" 
        << gkName << '"' << endl;
      return FALSE;
    }
  }
  else if (!args.HasOption('n')) {
    cout << "Searching for gatekeeper..." << flush;
    if (DiscoverGatekeeper(new H323TransportUDP(*this)))
      cout << "\nGatekeeper found: " << *gatekeeper << endl;
    else {
      cout << "\nNo gatekeeper found." << endl;
      if (args.HasOption("require-gatekeeper")) 
        return FALSE;
    }
  }
  
  return TRUE;
}


void COH323EndPoint::OnConnectionEstablished(H323Connection & connection,
                                             const PString & token)
{
  coutMutex.Wait();
  totalEstablished++;
  cout << "Started: \"" << connection.GetRemotePartyName()
       << "\" " << token
       << " active=" << connectionsActive.GetSize()
       << " total=" << totalEstablished
       << endl;
  coutMutex.Signal();
}


void COH323EndPoint::OnConnectionCleared(H323Connection & connection,
                                         const PString & token)
{
  coutMutex.Wait();
  cout << "Cleared: \"" << connection.GetRemotePartyName()
       << "\" " << token
       << " reason=" << connection.GetCallEndReason()
       << endl;
  coutMutex.Signal();
}


BOOL COH323EndPoint::OpenAudioChannel(H323Connection & /*connection*/,
                                      BOOL isEncoding,
                                      unsigned /* bufferSize */,
                                      H323AudioCodec & codec)
{
  codec.SetSilenceDetectionMode(H323AudioCodec::NoSilenceDetection);

  if (codec.IsRawDataChannelNative())
    return FALSE;

  if (!isEncoding)
    return FALSE;

  PWAVFile * wavFile = new PWAVFile("ogm.wav", PFile::ReadOnly);
  if (!wavFile->IsOpen()) {
    PTRACE(1, "CGEN323\tError opening ogm.wav: " << wavFile->GetErrorText());
    delete wavFile;
    return FALSE;
  }

  PDelayChannel * ogm = new PDelayChannel(PDelayChannel::DelayReadsOnly, 30);
  ogm->Open(wavFile, TRUE);
  codec.AttachChannel(ogm, FALSE);
  return TRUE;
}


///////////////////////////////////////////////////////////////////////////////

CallGen323::CallGen323()
  : PProcess("Callgen323", "Equivalence", MAJOR_VERSION, MINOR_VERSION, BUILD_TYPE, BUILD_NUMBER) 
{
}


void CallGen323::Main()
{
  PArgList & args = GetArguments();
  args.Parse("g-gatekeeper:"
             "i-interface:"
             "l-listen."
             "m-max:"
             "n-no-gatekeeper."
             "o-output:"
             "r-repeat:"
             "t-trace."
             "u-user:"
             "-tmincall:"
             "-tmaxcall:"
             "-tminwait:"
             "-tmaxwait:"
             "-require-gatekeeper."
              "-listenport:"         "-no-listenport."
             , FALSE);
  
  if (args.GetCount() == 0 && !args.HasOption('l')) {
    cout << "Usage:\n"
            "  callgen323 [options] -l\n"
            "  callgen323 [options] dest-host\n"
            "where options:\n"
            "  -l                   Passive/listening mode.\n"
            "  -m --max num         Maximum number of simultaneous calls\n"
            "  -r --repeat num      Repeat calls n times\n"
            "  -t --trace           Trace enable (use multiple times for more detail)\n"
            "  -o --output file     Specify filename for trace output [stdout]\n"
            "  -g --gatekeeper host Specify gatekeeper host\n"
            "  -i --interface addr  Specify IP address to bind to\n"
            "  -u --user username   Specify local username\n"
            "  -n --no-gatekeeper   Disable gatekeeper discovery [false]\n"
            "  --require-gatekeeper Exit if gatekeeper discovery fails\n"
            "  --listenport port    Specify the port of the listener\n"
            "  --tmincall secs      Minimum call duration in seconds [30]\n"
            "  --tmaxcall secs      Maximum call duration in seconds [30]\n"
            "  --tminwait secs      Minimum interval between calls in seconds [30]\n"
            "  --tmaxwait secs      Maximum interval between calls in seconds [30]\n"
            "\n";
    return;
  }
  
#if PTRACING
  PTrace::Initialise(args.GetOptionCount('t'),
                     args.HasOption('o') ? (const char *)args.GetOptionString('o') : NULL,
		     PTrace::Blocks | PTrace::DateAndTime | PTrace::Thread | PTrace::FileAndLine);
#endif
  
  COH323EndPoint endPoint;
  if (!endPoint.Initialise(args))
    return;
  
  if (args.HasOption('l'))
    cout << "Endpoint is listening for incoming calls.\n";
  else {
   
    unsigned number = args.HasOption('m') ? args.GetOptionString('m').AsUnsigned() : 1;
    unsigned repeat = args.HasOption('r') ? args.GetOptionString('r').AsUnsigned() : 10;
    unsigned tmin_call = args.GetOptionString("tmincall", "30").AsUnsigned();
    unsigned tmax_call = args.GetOptionString("tmaxcall", "30").AsUnsigned();
    unsigned tmin_wait = args.GetOptionString("tminwait", "30").AsUnsigned();
    unsigned tmax_wait = args.GetOptionString("tmaxwait", "30").AsUnsigned();

    if (tmin_call == 0 || tmin_wait == 0 || tmin_call > tmax_call || tmin_wait > tmax_wait) {
      cerr << "Invalid times entered!\n";
      return;
    }

    cout << "Endpoint starting " << number << " calls ";
    if (repeat != 0)
      cout << repeat;
    else
      cout << "infinite";
    cout << " times.\n";

    // create some threads to do calls, but start them randomly
    for (unsigned i = 0; i < number; i++)
      threadList.Append(new CallThread(i+1, endPoint, args[i % args.GetCount()], repeat, tmin_call, tmax_call, tmin_wait, tmax_wait));
  }

  coutMutex.Wait();
  cout << "Press ENTER at any time to quit" << endl;
  coutMutex.Signal();

  // wait for a keypress
  cin.get();

  coutMutex.Wait();
  cout << "\nAborting all calls ..." << endl;
  coutMutex.Signal();
  
  // stop threads and wait for cleanup
  while (threadList.GetSize() > 0) {
    threadList[0].Stop();
    threadList[0].WaitForTermination();
    threadList.RemoveAt(0);
  }

  // stop all calls
  endPoint.ClearAllCalls();
}


///////////////////////////////////////////////////////////////////////////////

CallThread::CallThread(unsigned _index,
                       COH323EndPoint & _endPoint,
                       const PString & _destination,
                       unsigned _repeat,
                       unsigned _tmin_call,
                       unsigned _tmax_call,
                       unsigned _tmin_wait,
                       unsigned _tmax_wait)
  : PThread(1000, NoAutoDeleteThread),
    endPoint(_endPoint),
    destination(_destination),          
    index(_index),
    repeat(_repeat),
    tmin_call(_tmin_call),
    tmax_call(_tmax_call),
    tmin_wait(_tmin_wait),
    tmax_wait(_tmax_wait)
{
  Resume();
}


void CallThread::Main()
{
  PTRACE(2, "Main\tStarted thread " << index);

  PRandom rand(PRandom::Number());

  // Loop "repeat" times for (repeat > 0), or loop forever for (repeat == 0)
  for (unsigned count = 1; repeat == 0 || count < repeat; count++) {

    // wait for a random delay
    PTimeInterval delay(((rand.Generate() % (tmax_wait - tmin_wait + 1)) + tmin_wait) * 1000);
    coutMutex.Wait();
    cout << setw(2) << index << ": Delaying for " << delay << " secs" << endl;
    coutMutex.Signal();

    PTRACE(1, "Main\tDelaying for " << delay);
    // wait for a random time
    if (exit.Wait(delay))
      break;

    // trigger a call
    PString token;
    PTRACE(1, "Main\tMaking call to " << destination);
    if (!endPoint.MakeCall(destination, token))
      PError << setw(2) << index << ": MakeCall to " << destination << " failed" << endl;
    else {
      delay = (((rand.Generate() % (tmax_call - tmin_call + 1)) + tmin_call) * 1000);
      coutMutex.Wait();
      cout << setw(2) << index << ": Making call " << count;
      if (repeat)
        cout << " of " << repeat;
      cout << " to " << destination << ' ' << token << " for " << delay << " secs" << endl;
      coutMutex.Signal();

      // wait for a random time
      PTRACE(1, "Main\tWaiting for " << delay);
      BOOL stopping = exit.Wait(delay);

      // end the call
      endPoint.ClearCall(token);

      if (stopping)
        break;
    }
  }

  coutMutex.Wait();
  cout << setw(2) << index << ": Completed call set." << endl;
  coutMutex.Signal();
  PTRACE(2, "Main\tFinished thread " << index);
}


void CallThread::Stop()
{
  coutMutex.Wait();
  cout << setw(2) << index << ": Stopping." << endl;
  coutMutex.Signal();
  exit.Signal();
}


// End of file ////////////////////////////////////////////////////////////////
